#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "quat.h"
#include "vmdsock.h"
#include "imd.h"
#include "nrutil.h"

#include "callback.h"
#include "powell.h"


#define xyz(cmd) { \
	        int     xyz; \
	        for (xyz=0; xyz<3; xyz++) cmd; \
}

#define dotp3(a,b) ((a)[0]*(b)[0]+(a)[1]*(b)[1]+(a)[2]*(b)[2])

static float orig[][3] = {
	{-0.853,   0.034,  -1.200 },
	{ -1.453,   0.098,   0.214 },
	{ -0.548,  -0.666,   1.193 },
	{ 0.853,  -0.034,   1.200 },
	{ 1.453,  -0.098,  -0.214 },
	{ 0.548,   0.666,  -1.193 },
};

int	port = 5000;

int	oneatom = 0;

struct cycle {
	float	(*orig)[3];
	int	n;
	float	dist[6],dotp[6];
	void	*vmd_conn;
	int	fix1;
	float	c1[3];
}; 

static void calc_dist_dotp(float atom[][3],int n,float dist[],float dotp[])
{
	int	i;
	int	ii;
	float	*a1,*a2,*a3;
	float	d,v[3],v2[3];

	for (i=0; i<n; i++) {
/* a1: atom i */
		a1 = atom[i];

/* a2: atom i+1 po cyklu */
		ii = (i+1) % n;
		if (ii >= n) ii -= n;
		a2 = atom[ii];

/* vektor a1-a2 */
		xyz(v[xyz] = a2[xyz]-a1[xyz]);
/* jeho delka */
		d = 0;
		xyz(d += v[xyz]*v[xyz]);
		dist[i] = sqrt(d);
	}

	for (i=0; i<n; i++) {
/*a1, a2: stejne
 *a3: atom i-1 po cyklu
 */
		a1 = atom[i];

		ii = (i+1) % n;
		if (ii >= n) ii -= n;
		a2 = atom[ii];

		ii = (i-1) % n;
		if (ii < 0) ii += n;
		a3 = atom[ii];

/* vektory a1-a2, a1-a3 */
		xyz(v[xyz] = a2[xyz]-a1[xyz]);
		xyz(v2[xyz] = a3[xyz]-a1[xyz]);

/* normalizace */
		xyz(v[xyz] /= dist[i]);
		xyz(v2[xyz] /= dist[ii]);

/* skalarni soucin */
		d = 0;
		xyz(d += v[xyz]*v2[xyz]);
		dotp[i] = d;
	}
}

static void rot_c612(float orig[][3],float rot,float shift,float c1[])
{
	float	*a1 = orig[0],*a2 = orig[1], *a6=orig[5];
	float	b[3],ax[3];
	float	alpha, d;
	q_type	qrot;
	q_vec_type	v;

/* pata kolmice z a1 na a6-a2 */

	alpha = - (dotp3(a1,a2)-dotp3(a2,a2)-dotp3(a1,a6)+dotp3(a2,a6)) /
		(-2*dotp3(a2,a6)+dotp3(a2,a2)+dotp3(a6,a6));

	xyz(b[xyz] = alpha*a6[xyz] + (1-alpha)*a2[xyz]);

/* normovana osa rotace */
	xyz(ax[xyz] = a2[xyz] - a6[xyz]);

	d = dotp3(ax,ax);
	d = 1/sqrt(d); 		/* dejme kompilatoru sanci */
	xyz(ax[xyz] *= d);


/* kvaternion rotace */
	q_make(qrot,ax[0],ax[1],ax[2],rot);

/* vektor b-a1 */
	xyz(v[xyz] = a1[xyz] - b[xyz]);

/* rotace b-a1 */
	q_xform(v,qrot,v);

/* nova poloha */
	xyz(c1[xyz] = b[xyz] + v[xyz] + shift*ax[xyz]);
}

/* vahy penalty za zmenu vzdalenosti a uhlu atomu */
#define W_DIST	1.5
#define W_DOTP	1

static float chex_min(float powell_vars[],void *arg)
{
	struct cycle	*c = arg;
	float	ret = 0;
	int	i;

	float	atoms[6][3], dist[6], dotp[6];

/* povodni polohy atomu - fixace */
	memcpy(atoms, c->orig, sizeof atoms);

/* pohyblive atomy */
	if (c->fix1) {
		xyz(atoms[0][xyz] = c->c1[xyz]);
	}
	else {
		xyz(atoms[0][xyz] = powell_vars[7+xyz]);
	}
	xyz(atoms[1][xyz] = powell_vars[1+xyz]);
	xyz(atoms[5][xyz] = powell_vars[4+xyz]);

/* vzdalenosti a uhly mezi novymi polohami */
	calc_dist_dotp(atoms,c->n,dist,dotp);

	for (i=0; i<6; i++) {
		float	d = dist[i] - c->dist[i];
		ret += W_DIST*d*d;

		d = dotp[i] - c->dotp[i];
		ret += W_DOTP*d*d;
	}

	usleep(20000);
	putchar('.'); fflush(stdout);
	imd_send_fcoords(c->vmd_conn,c->n,(float *) atoms);

	return ret;
}

void powell_cb(enum cb_action act,void *cb_arg)
{
	switch (act) {
		case CB_ACTION_ITER:
			printf("\nIterace\n");
			/* getchar(); */
			break;
		case CB_ACTION_DIR:
			printf("\nSmer\n");
			break;
		case CB_ACTION_LINSTART:
			printf("\n1D: ");
			break;
		case CB_ACTION_LINSTOP:
			printf("\n");
			break;
	}
}

int main(int argc,char **argv)
{
	void	*vmd_listen;
	struct cycle	ch;
	int	i,j,c;

/* XXX: indexy od 1 */
	float	**powell_dirs;
	float	powell_vars[10];	

	while (EOF != (c = getopt(argc,argv,"1"))) switch (c) {
		case '1': oneatom = 1; break;
	}

	ch.orig = orig;
	ch.n = 6;

/* napocitani vzdalenosti mezi atomy a uhlu */
	calc_dist_dotp(orig,ch.n,ch.dist,ch.dotp);

/* vytvoreni socketu a cekani na VMD */
	vmdsock_init();
	vmd_listen = vmdsock_create();
	vmdsock_bind(vmd_listen,port);

	printf("Cekam na portu %d\n",port);
	vmdsock_listen(vmd_listen);

	ch.vmd_conn = vmdsock_accept(vmd_listen);
	imd_handshake(ch.vmd_conn);

	while (1) {
		float	rot = 0,shift = 0;
		float	atoms[ch.n+1][3];
		char	buf[1000];

		int	iter;
		float	minret;

		printf("Rotace C1 (deg): ");
		fflush(stdout);
		scanf("%f",&rot); getchar();
		rot *= M_PI/180;

		printf("Posun C1 (A): ");
		fflush(stdout);
		scanf("%f",&shift); getchar();

/* inicializace pracovnich promennych
 * XXX: indexujeme od 1 kvuli NRC */
		memcpy(&atoms[1],orig,sizeof orig);

/* preklopeni uhliku do vysinute polohy */
		rot_c612(orig,rot,shift,ch.c1);
		xyz(atoms[1][xyz] = ch.c1[xyz]);

/* odeslani na VMD */
		imd_send_fcoords(ch.vmd_conn,ch.n,atoms[1]);
		printf("Pocatecni poloha odeslana\n");
		getchar();


/* inicializace smeru pro Powellovu metodu jednotkovymi vektory */
		powell_dirs = matrix(1,9,1,9);
		for (i=1; i<=9; i++) 
			for (j=1; j<=9; j++) 
				powell_dirs[i][j] = i==j ? 0.25 : 0;

/* inicializace pocatecnich hodnot 
 * XXX: souradnice atomu 6,1,2 do promennych 1-9
 * XXX: indexovani od 0 vs. 1
 * */

		ch.fix1 = 1;
		xyz(powell_vars[1+xyz] = orig[1][xyz]);
		xyz(powell_vars[4+xyz] = orig[5][xyz]);

/* minimalizace */

		powell(powell_vars,powell_dirs,6,1e-1,&iter,&minret,
				chex_min,powell_cb,&ch);

		printf("Hotovo, iter = %d, ret = %f\n\n",iter,minret);
		getchar();

/* totez podruhe s uvolnenym prvnim atomem */
		for (i=1; i<=9; i++) 
			for (j=1; j<=9; j++) 
				powell_dirs[i][j] = i==j ? 0.25 : 0;

/* inicializace pocatecnich hodnot 
 * XXX: souradnice atomu 6,1,2 do promennych 1-9
 * XXX: indexovani od 0 vs. 1
 * */

		ch.fix1 = 0;
		xyz(powell_vars[1+xyz] = orig[1][xyz]);
		xyz(powell_vars[4+xyz] = orig[5][xyz]);
		xyz(powell_vars[7+xyz] = ch.c1[xyz]);

/* minimalizace */

		powell(powell_vars,powell_dirs,9,1e-1,&iter,&minret,
				chex_min,powell_cb,&ch);

	}
}
